"""
$RCSfile: CVSFile.py,v $

The CVSFile package is a Zope product.  CVSFile allows one to use
CVS as a version control system for Zope content, by pointing to a
file in the developer's CVS sandbox on the Zope server.

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/02/18 19:20:04 $'[6:-2]
__version__ = '$Revision: 1.53 $'[10:-2]

# Base classes, peer classes
from Products.ExternalFile.ExternalFile import ExternalFile
import Products.ExternalFile.ExternalFile
from ICVSFile import ICVSFile
from CVSSandboxRegistry import CVSSandboxRegistry
from SandboxInfo import SandboxInfo

# Zope builtins
from Globals import DTMLFile, MessageDialog
from AccessControl import ClassSecurityInfo
import Globals  # InitializeClass (security stuff)

# Python builtins
from cStringIO import StringIO
import errno, os, os.path, socket, string, sys

################################################################
# Utility functions
################################################################

command_separator       = ';'
if sys.platform == 'win32':
    command_separator = '&'

def runCommand(command):
    """

    send passed in commandline to OS, return output or '(no
    output)' if there is no output

    """
    #print "running this command", command
    input, output_and_error  = os.popen4(command)
    output = output_and_error.read()
    if not output:
        output = "(no output)"
    return output

################################################################
# Container Methods
################################################################

# globals() is standard python for passing the current namespace
manage_addCVSFileForm       = DTMLFile('dtml/create',                      globals())
create_cvsregistry_formpart = DTMLFile('dtml/create_cvsregistry_formpart', globals())
create_cvssandbox_formpart  = DTMLFile('dtml/create_cvssandbox_formpart',  globals())

def findCVSSandboxRegistry(self):
    """

    Find an instance of CVSSandboxRegistry to obtain sandbox
    information.
    
    All CVSFile instances require an instance of CVSSandboxRegistry.
    The current policy is as follows: use the first registry found
    by acquisition.  Alternatively, a more complex policy could be
    used such as 
    a) looking in some predefined place, where the place could be 
    specified in some property (a la Zope3 explicit acquisition), 
    or 
    b) finding all registries in the ZODB and presenting a list for 
    selection by the user.
    For right now, we don't want to over-engineer the damn thing. 
    CKS 2/14/2002.
      
    """
    # Note: superValues does _NOT_ find things in the "Root Folder"
    registries = self.superValues('CVS Sandbox Registry')
    if registries:
        return registries[0]
    else:
        return None

def manage_addCVSFile(self,
                      id='', title='', description='',                  # standard
                      target_filepath='', basedir='', upload_file=None, # externalfile
                      REQUEST=None):
    """
    
    Add a CVSFile to a folder.  Called in response to create.dtml
    form.  Calls create_externalfile to do the work of confirming
    actions, creating the target file on the server if necessary, and
    finally calling our create_object method.


    """
    # make sure we got a file
    if not target_filepath:
        return MessageDialog(title   = 'ERROR',
                             message = 'You must specify a file.',
                             action  = 'manage_main')

    # The filepath is specified relative to a sandbox.  That way each
    # developer can use a different sandbox.  By prepending their
    # sandboxdir, we point at their version of the file

    return Products.ExternalFile.ExternalFile.manage_add(self,
                                                         id, title, description,                # standard
                                                         target_filepath, basedir, upload_file, # externalfile
                                                         REQUEST)

def create_file(self, contents_file,
                basedir, target_filepath,
                REQUEST):
    """
    
    Creates a physical file on the filesystem at basedir +
    targetfilepath.  If Contents_file is not None, it points to a file
    whose contents will be copied into the targetfile.  Otherwise, a
    zero length file will be created.  This method will raise an
    exception if there are any problems writing to the targetfile (or
    reading contents_file).

    """

    # first create the file
    create_msg = Products.ExternalFile.ExternalFile.create_file(self, contents_file,
                                                                basedir, target_filepath,
                                                                REQUEST)
    # then if we have gotten this far do a CVS add
    registry = self.findCVSSandboxRegistry()
    if not registry:
        return MessageDialog(title   = 'ERROR',
                             message = 'CVS Sandbox Registry not found.',
                             action  = 'manage_main')
    si = registry.getCurrentSandbox(REQUEST)    

    # According to the cvs manual, you must physically "be" inside the
    # directory to do a cvs add, you cannot traverse down directories.
    # That is: cd /home/cstrong/sandbox; cvs add foo/bar/mumble.txt BAD
    # cd /home/cstrong/sandbox/foo/bar; cvs add mumble.txt GOOD
    # See docs: http://www.cvshome.org/docs/manual/cvs_7.html
    # CKS 2/18/2002

    dir      = basedir + os.path.dirname(target_filepath)
    filename = os.path.basename(target_filepath)
    
    command  = "cd %s %s cvs -d %s add %s" % ( dir, command_separator,
                                               si.cvsroot, filename )

    return create_msg + runCommand(command)

def create_object(self, id, title,
                  description, basedir,
                  target_filepath,
                  REQUEST):
    """
    
    Factory method that actually creates the Zope object.  Assumes
    that all necessary checks and confirmations have been done.
    Assumes that target_filepath points to a valid, existing file with
    appropriate permissions in the filesystem/CVS sandbox in question.

    """
    self._setObject(id, CVSFile(id, title,
                                description,
                                target_filepath))
    self._getOb(id).reindex_object()

################################################################
# CVSFile class
################################################################

class CVSFile(ExternalFile):

    """
    Extended from ExternalFile
    """

    meta_type = 'CVS File'  # This is the name Zope will use for the Product in
                            # the "addProduct" list

    __implements__ = ICVSFile

    # This tuple defines a dictionary for each tab in the management interface
    # label = label of tab, action = url it links to
    manage_options = Products.ExternalFile.ExternalFile.ExternalFile.manage_options + (
	{'label':'CVS',	    'action': 'manage_cvsForm',     'help': ('CVSFile', 'cvs.stx')     },
        )

    _security = ClassSecurityInfo()

    # set security for the object itself, e.g. if it is accessed in DTML code
    # This line is REQUIRED to allow access to CVSFiles by the public (ariel DTDs, etc.)
    _security.declareObjectPublic()

    # call __call__() method on DTMLFile for our edit form, point at our cvs_form.dtml
    # globals() is standard python for passing the current namespace
    cvs_form           = DTMLFile('dtml/cvs_form',globals())
    
    _security.declareProtected('Manage CVS Files',      'manage_cvsForm')
    manage_cvsForm     = DTMLFile('dtml/cvs',globals())

    # these three exist so we can call them from DTML
    create_cvsregistry_formpart = create_cvsregistry_formpart
    create_cvssandbox_formpart  = create_cvssandbox_formpart
    # end of DTML scoping hack

    ################################################################
    # Methods implementing the ICVSFile interface below
    ################################################################

    _security.declareProtected('View',  'findRegistry')
    def findRegistry(self):
        """

        Find an instance of CVSSandboxRegistry to obtain sandbox
        information.  This calls a module scoped function which does
        the real work of finding the sandbox.  Because we need this
        behavior to even _create_ an instance of CVSFile the function
        needs to be module scoped.

        """
        
        return findCVSSandboxRegistry(self)

    _security.declareProtected('View',  'getCurrentSandbox')
    def getCurrentSandbox(self, REQUEST):
        """

        Convenience function for getting the current sandbox from the
        sandbox registry

        """
        registry = self.findRegistry()
        if not registry:
            raise EnvironmentError, 'CVS Sandbox Registry not found'
        else:
            return registry.getCurrentSandbox(REQUEST)    


    def updateFile(self, si):
        """

        Perform a 'cvs update' on the CVS file pointed to by this
        object, using the sandbox information from the SandboxInfo
        parameter.

        """
        command = "cd %s %s cvs -d %s -q update %s" % ( si.basedir,
                                                        command_separator,
                                                        si.cvsroot,
                                                        self.filepath )
        return runCommand(command)

    def revertFile(self, filepath, si):
        """

        Remove the sandbox version of the file pointed to by this
        object, the do a 'cvs update' and pass the output back to the
        form.  This is used to throw away any non-checked in changes
        to the file.  Retrieve sandbox information from the passed-in
        SandboxInfo parameter.

        """

        # remove sandbox version of file
        os.remove(filepath)

        # then do a cvs update
        return self.updateFile(si)

    def commitFile(self, si, commitMessage, userName):
        """

        Perform a commit on the CVS file pointed to by this object.
        This will submit local changes to the central repository.
        Retrieve sandbox information from the passed-in SandboxInfo
        parameter.  Include the passed in username in the commit
        message, because otherwise there would be no way to
        distinguish whodunit, because all userids are the same, namely
        the userid under which the Zope server was started.
        
        """
        who = 'Zope CVSFile user: %s\n' % userName

        command = "cd  %s %s cvs -d %s commit -m \"%s\" %s" % ( si.basedir,
                                                                command_separator,
                                                                si.cvsroot,
                                                                who + commitMessage,
                                                                self.filepath )
        return runCommand(command)

    ################################################################
    # Methods implementing the IExternalFile interface
    ################################################################

    _security.declareProtected('View',  'getFilepath')
    def getFilepath(self, REQUEST):
        """
        
        Get the full path of the external file this CVSFile is
        pointing to.  Depends on the sandbox we are currently using.
        
        """
        si = self.getCurrentSandbox(REQUEST)
        return si.basedir + self.filepath

    ################################################################
    # Methods called from the CVS form 
    ################################################################

    # called from cvs.dtml form
    _security.declareProtected('Manage CVS Files',      'cvsUpdateFile')
    def manage_UpdateFile(self, REQUEST):
        """
        Does a CVS update on the file and passes output back to the form.
        """
        si = self.getCurrentSandbox(REQUEST)
        output = self.updateFile(si)
        return self.manage_cvsForm(updateFileOutput = output)    


    # called from cvs.dtml form
    _security.declareProtected('Manage CVS Files',      'cvsRevertFile')
    def manage_RevertFile(self, REQUEST):
        """
        
        Removes the sandbox version of a file, then does a CVS update
        on the file and passes output back to the form.  This, in
        effect, throws away any non-checked in changes to the file.

        """
        filepath = self.getFilepath(REQUEST)
        si       = self.getCurrentSandbox(REQUEST)
        output   = self.revertFile(filepath, si)
        return self.manage_cvsForm(revertFileOutput = output)

    # called from cvs.dtml form
    _security.declareProtected('Manage CVS Files',  'cvsCommitFile')
    def manage_CommitFile(self, commitMessage='', userName='',
                          REQUEST=None):
        """
        Does a CVS commit on the file and passes output back to the form
        """
        #
        # Since Zope always runs as an administrative user
        # (e.g. "nobody") we won't know who actually committed these
        # changes unless we capture it another way.  Since by
        # convention we make the Zope usernames the same as the
        # UNIX/CVS usernames, we grab it from Zope via a hidden field
        # in the cvs.dtml form.
        #
        if not commitMessage:
            return MessageDialog(title   = 'ERROR',
                                 message = 'You must specify a commit message.',
                                 action  = 'manage_cvsForm')
        
        si     = self.getCurrentSandbox(REQUEST)
        output = self.commitFile(si, commitMessage, userName)
        return self.manage_cvsForm(commitFileOutput = output)

    ################################################################
    # Support for Schema Migration
    ################################################################

    def repair(self):
        """

        Repair this object.  This method is used for schema migration,
        when the class definition changes and existing instances of
        previous versions must be updated.

        """
        if not hasattr(self, 'behave_like'):
            self.behave_like = 'DTMLDocument'

# register security information
Globals.InitializeClass(CVSFile)
